#include "Arduino.h";
#include "dmxclient.h";
#include "EEPROM.h";


uint8_t value = 0;
uint16_t arraypos;
uint16_t wait;
bool retbool = false;

/*********************************************************************
	grab a pointer to the class receiver..
**********************************************************************/
DMXClientClass *dmxlibclassptr;

/*********************************************************************
	use the pointer to the class receiver in the ISR for the UART receiver
**********************************************************************/
ISR(USART_RX_vect) {
	dmxlibclassptr->receiver();
}

/*********************************************************************
	initialise the DMX receiver
**********************************************************************/
void DMXClientClass::begin(uint8_t baseChannel, uint8_t channelCount, bool forceChannel, uint8_t ledPin, uint8_t setButtonPin, uint16_t eepromAddress, DMXMODE dmxMode) {

	dmxlibclassptr = this;  // !! important to set for the ISR!
	
	this->eepromAddress = eepromAddress;
	
	this->dmxAddress = EEPROM.read(this->eepromAddress);
	if ((dmxAddress == 0xFF) || (dmxAddress == 0) || (forceChannel)) {
		dmxAddress = baseChannel;
		EEPROM.update(eepromAddress,dmxAddress);
	}
	
	dmxChannelCount = EEPROM.read(eepromAddress + 1);
	if ((dmxChannelCount == 0xFF) || (dmxChannelCount < 10) || (forceChannel)) {
		dmxChannelCount = channelCount;
		EEPROM.update(eepromAddress + 1,dmxChannelCount);
	}

	this->ledPin = ledPin;
	if (ledPin > 0) {
		pinMode(ledPin, OUTPUT);	
		digitalWrite(ledPin,LOW);
	}
	
	this->setButtonPin = setButtonPin;
	if (setButtonPin > 0) {
		pinMode(setButtonPin,INPUT_PULLUP);
	}
	
	this->dmxBuffer = (uint8_t*) malloc(this->dmxChannelCount + 3);
	
	for (uint8_t genp=0; genp < this->dmxChannelCount; genp++) {
  	this->dmxBuffer[genp] = 0;								// set our data to 0
  }
  
  this->dmxData = false;
  this->dmxMode = dmxMode;
  
  if (dmxMode == DMXCONT) {
  	this->setMode(UARTFULL);
  	
  } else if (dmxMode == DMXPOLL) {
  	this->setMode(UARTFULL);
  	
  } else {
  	this->setMode(UARTOFF);
  }
}

/*********************************************************************
		set our DMX mode to off, readonly or read and run interrupts
**********************************************************************/
void DMXClientClass::setMode(UARTMODE uartMode = UARTMODE::UARTRONLY) {
	switch(uartMode) {
		case UARTMODE::UARTOFF :
			UCSRnB = 0;
			break;
		
		case UARTMODE::UARTRONLY :
			UBRRnH = dmxPreScale >> 8;
    	UBRRnL = dmxPreScale;
    	//UCSRnB = (1 << RXENn);
    	UCSRnB = 0;
    	UCSRnC = SERIAL_8N2;	
    	break;
    	
    case UARTMODE::UARTFULL :
    	UBRRnH = dmxPreScale >> 8;
    	UBRRnL = dmxPreScale;
    	UCSRnB = (1 << RXENn) | (1 << RXCIEn);
    	UCSRnC = SERIAL_8N2;	
    	break;
	}
	
}
	

/*********************************************************************
	DO NOT CALL THIS FUNCTION YOURSELF! IT IS AN INTERRUPT ROUTINE! 
	called by the ISR set up earlier. 
**********************************************************************/
void DMXClientClass::receiver(void) {
	uint8_t recvError = (UCSRnA & (1 << FEn)); 	// get state before data!
  uint8_t recvData = UDRn; 										// get data
  uint8_t localState = this->bufferState;			// get the state only once from global to save time.
  uint16_t localCountdown = this->dmxCountdown;
  
  if (localState == RECVSTART) {
  	this->bufferState = RECVIDLE;
  	return;
  }
  
  if (recvError) {
  	this->bufferState = RECVSKIPSTARTBYTE;
  	this->dmxCountdown = this->dmxAddress;
  	this->dmxCurrChannel = 0;
  	this->dmxBuffPos = 0;
  	this->lastDmxPacketTime = millis();
  	
  } else {
  	switch (localState) {
  		
  		case RECVSKIPSTARTBYTE :
  			if (recvData == 0) {		// if that first byte is not 0, its some other dmx thing and we ignore it.
  				this->bufferState = RECVSAVEDATA;
  				this->dmxDataChanged = false;
  			} else {
  				this->bufferState = RECVIDLE;
  			}
  			break;
  	
  		case RECVSAVEDATA :
  			dmxCurrChannel++;
  			
  			if (setState) {
  				// we're setting the base channel here..
  				
  				if (channelScanCount > 15) {
  					setState = false;
  					this->dmxAddress = lastChannelScan;
  					this->dmxSaveSetting = true;
  					this->dmxData = true;
  					this->bufferState = RECVIDLE;
  				} else {
  					if (recvData > 50) {
  						if (recvData != lastChannelScanVal) {
  							if (dmxCurrChannel != lastChannelScan) {
  								channelScanCount = 0;
  								lastChannelScan = dmxCurrChannel;
  							} else {
  								channelScanCount++;
  							}
  						}
  					}
 					}
  				
  				
  			} else {
  				if (this->dmxCountdown > 1) {
  					this->dmxCountdown--;
  				} else {
  					//if (this->dmxBuffer[this->dmxBuffPos] != recvData) {
  						this->dmxBuffer[this->dmxBuffPos] = recvData;
  						this->dmxDataChanged = true;
  					//}
  					this->dmxBuffPos++;
  				}
  			
  				if (this->dmxBuffPos >= this->dmxChannelCount) {
  					this->dmxData = true;
  					this->bufferState = RECVIDLE;
  					if (this->dmxMode == DMXMODE::DMXPOLL) {
  						this->setMode(UARTMODE::UARTRONLY);
  						this->bufferState = RECVDONE;
  					}
  				}
  			}
  			
  			break;
  	}
  }  
}


/*********************************************************************
	call this in the main loop if not using poll.
	poll will call this before doing anything else 
**********************************************************************/
void DMXClientClass::loop (void) {
	
	if (this->dmxSaveSetting) {
		EEPROM.update(eepromAddress,this->dmxAddress);
		this->dmxSaveSetting = false;
	}
	
	lastSeen();
	updateLed();
	checkSetButton();
	
	if (lastSeen()) {
		dmxState = true;		
	} else {
		dmxState = false;
	}
	
}


/*********************************************************************
	poll for one dmx packet
**********************************************************************/
bool DMXClientClass::poll(void) {
	retbool = false;
	
	loop();
	
	if (this->dmxMode == DMXMODE::DMXPOLL) {
		this->dmxData = false;
		this->bufferState = RECVSTART;
		this->setMode(UARTMODE::UARTFULL);
		while (UCSRnA & (1 << RXCn)) {
    	value = UDRn; // get data
  	}
  	wait = DMXPROBE_MAX_TIME;
  	while ((wait > 0) && (this->bufferState != RECVDONE)) {
      delay(1);
      wait--;
    } // while
    
  	if (this->bufferState == RECVDONE) {
  		retbool = true;
  	} else {
  		this->setMode(UARTMODE::UARTRONLY);
  	}
	}
	
	return retbool;
}
	
	
/*********************************************************************
	get the DMX data from the given channel.
**********************************************************************/
uint8_t DMXClientClass::readChannel(uint16_t channel) {
	arraypos = channel - this->dmxAddress;

	if (arraypos < 0) {
		arraypos = 0;
	} else if ( arraypos > this->dmxChannelCount) {
		arraypos = this->dmxChannelCount;
	}
	value = this->dmxBuffer[arraypos];
	
	return value;
}


/*********************************************************************
	update the led state	
**********************************************************************/
void DMXClientClass::updateLed() {
	
	if (ledPin > 0) {
		if (millis() - lastLedTime > ledTimeout) {
			this->ledState = !this->ledState;
			digitalWrite(this->ledPin,this->ledState);
			this->lastLedTime = millis();
			
			if (! setState) {
				if (ledState) {
					if (dmxData) {
						ledTimeout = 50;
					} else {
						ledTimeout = 300;
					}
				} else {
					if (extSwitch) {
						ledTimeout = 1000;
					} else {
						ledTimeout = 300;
					}
				}
			} else {
				if (ledState) {
					ledTimeout = 1000;
				} else {
					ledTimeout = 50;
				}
			}
		}
	}
}

/*********************************************************************
	lastseen: returns false if we never had a DMX signal or the last one was too long ago.
	Else, it will return true.
**********************************************************************/
bool DMXClientClass::lastSeen(void) {
	this->dmxData = false;
	
	if (this->lastDmxPacketTime == 0) {
		return false;
	
	} else if (millis() - this->lastDmxPacketTime > this->signalTimeout) {
		return false;
		
	}
	
	this->dmxData = true;
	return true;
}


/*********************************************************************
	update the led state	
**********************************************************************/
void DMXClientClass::checkSetButton(void) {
	if (setButtonPin > 0) {
		if (millis() - buttonDebounce > 100) {
			buttonCurrentState = digitalRead(setButtonPin);
			
			if (! buttonCurrentState) {
				if (buttonPrevState) {
					
					if (buttonTimeout == 0) {
						buttonTimeout = millis();
					} else {
						if (millis() - buttonTimeout > SETTIMEOUT) {
							buttonPrevState = false;
							if (! setState) {
								setState = true;
								channelScanCount = 0;
								lastChannelScan = 0;
								lastChannelScanVal = 0;
							} else {
								setState = false;
							}
						}
					}
				}
			} else {
				buttonPrevState = true;
				buttonTimeout = 0;
			}
			buttonDebounce = millis();
		}
	}
}