/*
		the main MDC client code. This handles everything to get the data from the line.
		
		It has a lot of optimizations for an ATTINY, where space is premium, but also it is not possible to assign
		any port numbers above port 5, meaning it can only ever use registers PORTB, DDRB and PINB.
		
		On arduino nano etc, this is more difficult to determine but has more memory, 
		so the lazy way of using the arduino functions to set, read and write to pins is used.
*/
		
#include "arduino.h"
#include "EEPROM.h"
#include "mdcclient.h"
#include "BlockTimes.h"
#include "mdcPkgManager.h"

#ifdef USEHAMMING
#include "HammingCodes.h"
#endif

#if defined(__AVR_ATtiny85__) || defined(__AVR_ATtiny45__)
#undef SERIALMON
#endif
/*****************************************************************************
*
* make a pointer to the instance of mdcClient, make sure to assign it 'this' in the setup.
*
******************************************************************************/
mdcClientClass *instancepointer;

/*****************************************************************************
*
* the global functions for interrupt callers
*
******************************************************************************/
void int0handlertrigger ( void ) {
	instancepointer->int0handler();
}


/*void int0scantrigger ( void ) {
	instancepointer->int0scanner();
}*/

#ifdef TINY

ISR(INT0_vect){
   instancepointer->int0handler();
}

#endif


/*****************************************************************************
*
* begin function to call.
*
******************************************************************************/
void mdcClientClass::begin(uint8_t baseChannel, uint8_t channelCountIn, bool channelForce, uint8_t ledPinIn, uint8_t setButtonPinIn, uint16_t eepromAddress) {

	instancepointer = this;  // !!!! IMPORTANT TO SET THIS THING!!
	
#ifdef SERIALMON
	Serial.begin(115200);
	Serial.println(F("Startin MDC client"));
#endif	
	
	ledState 						= false;
	setButtonState 			= false;
	this->ledPin				= ledPinIn;
	this->setButtonPin	= setButtonPinIn;
	this->channelCount  = channelCountIn;
	
	myBaseChannel = EEPROM.read(eepromAddress);
	
	if ((myBaseChannel == 0) || (channelForce == 1) || (myBaseChannel == 255)) {
		myBaseChannel = baseChannel;
		EEPROM.update(eepromAddress,myBaseChannel);
	}

#ifndef TINY
	if (this->ledPin > 0) {
		pinMode(this->ledPin,OUTPUT);
		digitalWrite(this->ledPin,LOW);
	}
	
	if (setButtonPin > 0) {
		pinMode(setButtonPin,INPUT_PULLUP);
	}
#endif	
	

// we can use the above for the attiny, but direct port assignment saves space.
// and we know a port cannot be > 5, in registers DDRB, PORTB and PINB
#ifdef TINY
	MCUCR |= ( 1<<ISC01 | 1<<ISC00 );			// ISC 11 = rising edge, 10 = falling edge, 01 = logic change.
  GIMSK |= 1<<INT0;  										// int0 interrupt
  // the actual connection is made earlier in ISR declaration.
  
  if (this->ledPin > 0) {
  	DDRB |= ( 1<<this->ledPin); // set ledpin port as output
  }
  if (setButtonPin > 0) {
  	DDRB &= ~(1<<setButtonPin);  // set buttonpin port as input
  	PORTB |= (1<<setButtonPin);	 // activate pull-ups on that port.
  }
  	
  
#else
	attachInterrupt(digitalPinToInterrupt(2), int0handlertrigger , RISING);  
	//	attachInterrupt(digitalPinToInterrupt(2), int0scantrigger, CHANGE);
	edgetrigger = RISING;
#endif	
	
	signalvalid = true;
	errorcount = 0;
	//validcount = 0;
	
#ifdef SERIALMON
	Serial.println(F("MDC client started"));
	Serial.print(F("Listening on channel: "));
	Serial.println(myBaseChannel);
#endif

}

/*****************************************************************************
*
* update function, it will return true if there was an update on any channel.
# call it in the loop as: if (mcdclient.update()) { ..... }
*
******************************************************************************/
bool mdcClientClass::update(void) {
	hasUpdates = false;
	
	/*if ((validcount > 5) && signalvalid) { // future use when i figured this out.
		Serial.println(F("errors detected, rescanning edge"));
		signalvalid = false;
		validcount = 0;
		attachInterrupt(digitalPinToInterrupt(2), int0scantrigger, CHANGE);
	}*/
	
	if (errorcount > 5) {
		swapTriggerEdge();
		errorcount = 0;
	}	

	updateLed();
	check_settingbutton();
	
	
	if (queuePtr != leftPtr) {
		
		if (++leftPtr == MAXQUEUE) {
			leftPtr = 0;
		}
		
		dataerror = false;
		
#ifdef USEHAMMING				
		queue[leftPtr].hammingChannelH = pgm_read_byte(hammingRX + queue[leftPtr].hammingChannelH);
		queue[leftPtr].hammingChannelL = pgm_read_byte(hammingRX + queue[leftPtr].hammingChannelL);
		queue[leftPtr].hammingValueH = pgm_read_byte(hammingRX + queue[leftPtr].hammingValueH);
		queue[leftPtr].hammingValueL = pgm_read_byte(hammingRX + queue[leftPtr].hammingValueL);
		
		dataerror = ( (queue[leftPtr].hammingChannelH > 0x0F) || (queue[leftPtr].hammingChannelL > 0x0F));
		dataerror |= ( (queue[leftPtr].hammingValueH > 0x0F) || (queue[leftPtr].hammingValueL > 0x0F));
		
		if (! dataerror) {
			mdcChannel = (queue[leftPtr].hammingChannelH << 4) + queue[leftPtr].hammingChannelL;		
		} else {
			mdcChannel = 255;
		}
		mdcValue = (queue[leftPtr].hammingValueH << 4) + queue[leftPtr].hammingValueL;
		
#else
		mdcChannel = queue[leftPtr].hammingChannelL;
		mdcValue = queue[leftPtr].hammingValueL;
#endif;	

#ifdef SERIALMON
		Serial.print(mdcChannel);
#endif		

		if (mdcChannel > 250) {
			errorcount++;
		}
		
		if (setButtonState) {
			check_channelset(mdcChannel, mdcValue);
			
		} else {

#ifdef SERIALMON				
				Serial.print(F(" = "));
				Serial.print(mdcValue);
#endif

			if ( (! dataerror) && (mdcChannel >= myBaseChannel) ) {
				localChannel = mdcChannel - myBaseChannel;

				if ( localChannel < channelCount) {
					hasUpdates = true;
					errorcount = 0;
#ifdef SERIALMON
					Serial.print(F(" = local "));
					Serial.print(localChannel);
				} else {
					Serial.print(F(" not for me"));
#endif					
				}
#ifdef SERIALMON
			} else {
				Serial.print(F(" not for me"));
#endif				
			}
		}
#ifdef SERIALMON		
		Serial.println();			
#endif
	}
	return this->hasUpdates;
}


/*****************************************************************************
*
* update the led state depending on some set parameters at other places.
*
******************************************************************************/
void mdcClientClass::updateLed(void) {
	if (ledPin > 0) {
		if (millis() - ledUpdate > ledTimeout) {
			ledState = !ledState;
			ledTimeout = LEDREFRESH;
			ledUpdate = millis();		
			
			if (!setButtonState && signalvalid) {
				if (millis() - lastInt0Update > 500) {
					ledMode = 0;
				}
				
				switch (ledMode) {
					case 0 :								 // no signal detected..
						ledState = false;
						break;
					//case 1 :								// some block wave detected, data is not recognized. just here for reference, it doesnt change anything.
					//	break;
					
					case 2 :								// block wave and valid data detected or scanning..
						if (ledState) {
							ledTimeout = 50;
						} else {
							ledTimeout = 1000;
						}
						break;
						
					//case 3 :							// interrupt detection method, not in use at the moment.
					//	ledTimeout = 100;
					//	break;				
						
				}
								
			} else {
			
				if (! ledState || ! signalvalid) {
					ledTimeout = 100;
				} else {
					ledTimeout = 500;
				}
			}
#ifdef TINY
			PORTB ^=(1<<ledPin);			
#else
			digitalWrite(ledPin,ledState);
#endif			
		}
	}

}


/*****************************************************************************
*
* the actual normal interrupt handler responsible for reading the bits and moving them.
*
******************************************************************************/
void mdcClientClass::int0handler(void) {
	lastInt0Update = millis();
	if (! ledMode) { ledMode = 1; }
		
	prevUsec = currentUsec;
	currentUsec = micros();
	timeDiff = currentUsec - prevUsec;
	validPackage = false;
	
	if (timeDiff < MAX_ONE_BIT) {
		bit = true;
		
	} else if (timeDiff < MAX_ZERO_BIT) {
		bit = false;
		
	} else if (timeDiff < MAX_SPACE_BIT) {
		// seems a space bit, so we set things to start assuming a start.
		// even when the next one is a 1 bit, the next one will again be a start.
		pkgState = PKGSTART;
	} else {
		pkgState = PKGIDLE;
	}
	
	switch (pkgState) {
		case PKGSTART :
			modeBits = 0;
			bitStream = 0;
			pkgState = PKGMODEBITS;
			intCounter = 2;
			break;
			
		case PKGMODEBITS :
			modeBits = (modeBits << 1) | bit;
			if (! --intCounter) {
				pkgState = PKGDATA;
				intCounter = BITS;
			}
			break;
			
		case PKGDATA :
			bitStream = (bitStream << 1) | bit;
			if (! --intCounter) {
				ledMode = 2;
				pkgState = PKGEND;
				if (bitStream > 0) {
					validPackage = true;
				}
			}
			break;
	}

	if (validPackage) {
		if (++queuePtr >= MAXQUEUE) {
			queuePtr = 0;
		}
		//memcpy(&queue[queuePtr].bitStream,&bitStream,sizeof(bitStream));
		queue[queuePtr].bitStream = bitStream; // this works fine, faster and with 12 less bytes than memcpy somehow..
		queue[queuePtr].modebits = modeBits;
	}
}


/*****************************************************************************
*
* the int0 scanner handler to determine the polarity of the stream.
*
******************************************************************************/
/*void mdcClientClass::int0scanner(void) {
	ledMode = 3;
	currentUsec = micros();
	udiff[udiffidx] = currentUsec - prevUsec;
	prevUsec = currentUsec;
	
	if (! skipme) {
		udiffidx ^= 0x01;
		if (abs(udiff[0] - udiff[1]) < 5) {
			validcount++;
		} else {
			skipme = true;
			
		}
		if (validcount > 20) {
			noInterrupts();
			if (digitalRead(2)) {
				edgetrigger = RISING;
			} else {
				edgetrigger = FALLING;
			}
			Serial.println(edgetrigger);
			attachInterrupt(digitalPinToInterrupt(2), int0handlertrigger, edgetrigger);
			signalvalid = true;
			validcount = 0;
			ledMode = 0;
			interrupts();
		}
	} else {
		skipme = false;
		validcount = 0;
	}
}*/


/*****************************************************************************
*
* check the setting button state and current value for scanning the channel to set.
*
******************************************************************************/
void mdcClientClass::check_settingbutton ( void  ) {
	if (setButtonPin > 0) {
		if (millis() - buttonDebounce > 100) {

#ifdef TINY
			buttonCurrentState = (PINB & (1<<setButtonPin));
#else					
			buttonCurrentState = digitalRead(setButtonPin);
#endif
		
			if (! buttonCurrentState) {
				if (buttonPrevState) {
					
					if (buttontimeout == 0) {
						buttontimeout = millis();
					} else {
						if (millis() - buttontimeout > SETTIMEOUT) {
							
							buttonPrevState = false;
							if (! setButtonState) {
								setButtonState = true;
								channelScanCount = 0;
								lastChannelScan = 0;
								lastChannelScanVal = 0;

#ifdef SERIALMON								
								Serial.println(F("setting channelset ON"));
#endif								
							} else {
								setButtonState = false;
#ifdef SERIALMON								
								Serial.println(F("setting channelset OFF"));
#endif
							}
						}
					}
				}
			} else {
				buttonPrevState = true;
				buttontimeout = 0;
			}
			buttonDebounce = millis();
		}
	}
}

/*****************************************************************************
*
* if scanning for a channel to set this is called whenever a channel update is received from the update function.
*
******************************************************************************/
void mdcClientClass::check_channelset (uint8_t newchannel, uint8_t newvalue) {

	if (newchannel != lastChannelScan) {
		channelScanCount = 0;
		lastChannelScanVal = 0;
		lastChannelScan = newchannel;	
	}
	
	
	if (newvalue != lastChannelScanVal) {
		lastChannelScanVal = newvalue;
		channelScanCount++;
	}
	
	if (channelScanCount > 10) {
		myBaseChannel = lastChannelScan;
		setButtonState = false;
		channelScanCount = 0;
		lastChannelScan = 0;
		lastChannelScanVal = 0;
		
		// set it to EEPROM.
		EEPROM.update(EEPROMADDR,myBaseChannel);

	}
}


/*****************************************************************************
*
* the global functions for interrupt callers
*
******************************************************************************/
void mdcClientClass::resetBuffer (void) {
	this->queuePtr = 0;
	this->leftPtr = 0;
}

/*****************************************************************************
*
* swap the interrupt trigger edge
*
******************************************************************************/
void mdcClientClass::swapTriggerEdge (void) {
	
	noInterrupts();
	
	if (edgetrigger == RISING) {
		
#ifdef TINY
		MCUCR |= ( 1<<ISC01 | 0<<ISC00 );			// ISC 11 = rising edge, 10 = falling edge, 01 = logic change.
#endif
		edgetrigger = FALLING;
	} else {
		
#ifdef TINY
		MCUCR |= ( 1<<ISC01 | 1<<ISC00 );			// ISC 11 = rising edge, 10 = falling edge, 01 = logic change.
#endif
		edgetrigger = RISING;
	}

#ifndef TINY	
	attachInterrupt(digitalPinToInterrupt(2), int0handlertrigger, edgetrigger);  
#endif

	interrupts();
}